/* imskkfep.c */					/* vi:set ts=4 sw=4: */
/*
 * IM for MGL using skkfep.
 *
 * Author:
 *	  KIHARA Hideto <deton@m1.interq.or.jp>
 *	  $Id: imskkfep.c,v 1.5 2000/01/28 11:07:58 deton Exp $
 */

#define FALSE			0
#define TRUE			(!FALSE)
#define NUL				'\0'
#define ESC				'\033'
#define PATHSEPSTR		"/"
#define Ctrl(c)			((c) & 0x1f)
#define MAX_COLUMNS		80
#define FEP_BUF_SIZE	((MAX_COLUMNS) * 2 + 1)
#define IsKanji(c)		((c) & 0x80)

static char fep_buf[FEP_BUF_SIZE] = "";
static char *fbhead = fep_buf, *fbtail = fep_buf;
static char delay_buf[FEP_BUF_SIZE] = "";
static int jpfep_flush = FALSE;
static int jpfep_mode = FALSE;
static int ready = FALSE;	/* initialized? */

static void fep_delchars();
static int fep_insstr();
static void update_textscreen();

#include "config.h"	/* skkfep's config.h */
#include "fep.h"
#include "functions.h"
KeymapPtr CurrentKeymap;
char UserDicName[64];
Dictionary UserDic = NULL;
#ifdef SYSTEM_DIC_NAME
Dictionary SystemDic;
#endif
#ifdef USE_SERVER
int ServerOK = 0;
#endif
#ifdef BOTH_SERVER_AND_SKKSRCH
int (*openSKKservPtr)() = openSKKserv;
CandList (*getCandFromServerPtr)() = getCandFromServer;
int (*closeSKKservPtr)() = closeSKKserv;
#endif
short BlockTty = 0;
char PreserveOnFailure = PRESERVE_ON_FAILURE;
char OkuriFirst = 0;
static int skkfep_lastmode = KANA_MODE;
static int skkfep_curmode = KANA_MODE;
static int skkfep_lastshowmode = KANA_MODE;

static char disp_buf[FEP_BUF_SIZE] = "";
static int disp_pos = 0;

	static int
jpfep_isready()
{
	return (fbhead < fbtail);
}

/* display SKK mode */
	void
showSKKmode(s)
	int s;
{

    skkfep_lastmode = skkfep_curmode;
    skkfep_curmode = s;
	switch (s) {
		case SKK_MODE:
		case KANA_MODE:
		case KKANA_MODE:
		case ZENEI_MODE:
		case CODEIN_MODE:
			if (s != skkfep_lastshowmode) {
				skkfep_lastshowmode = s;
				update_textscreen();	
			}
			break;
		default:	/* do not show other mode */
			break;
	}
}

	void
showlastmode()
{
    showSKKmode(skkfep_lastmode);
    skkfep_curmode = skkfep_lastmode;
}

/* temporary display some characters like candidates */
	int
writes(s)
	char *s;
{
	/* must flush out fep_buf before print s */
	/* ex:if typed "tabunDaijoubu", must not print "\/d" before flushOut "nn" */
	if (jpfep_flush && jpfep_isready()) {
		strncat(delay_buf, s, sizeof(delay_buf) - strlen(delay_buf) - 1);
	} else {
		jpfep_flush = FALSE;
		if (delay_buf[0] != NUL) {
			fep_insstr(delay_buf);
			delay_buf[0] = NUL;
		}
		fep_insstr(s);
	}
}

	int
write1(c)
	char c;
{
	char buf[2];
	buf[0] = c;
	buf[1] = NUL;
	writes(buf);
}

/* put character on fep_buf to return by jp_getchar() */
	int
writeShell1(c)
	char c;
{
#ifdef REGMODE
	if (Registering)
		regUpdateCand1(c);
	else
#endif /* REGMODE */
	if (fbtail >= fep_buf + FEP_BUF_SIZE - 1)
	{
		puts("Input buffer filled.");
		return 0;
	}
	else
	{
		*fbtail++ = c;
		*fbtail = NUL;
	}
	return 1;
}

/* put string on fep_buf to return by im_read() */
	int
writeShells(s)
	char *s;
{
#ifdef REGMODE
	if (Registering)
		regUpdateCands(s);
	else
#endif /* REGMODE */
	{
		strcpy(fbtail, s);	/* {{ must check overflow }} */
		fbtail += strlen(fbtail);
	}
	return strlen(s);
}

	int
msgLine()
{
	return 0;	/* {{ not yet support to use message line }} */
}

	void
toMsg()
{
}

	void
fromMsg()
{
}

	void
csrLeft(n)
	int n;
{
	disp_pos -= n;
	if (disp_pos < 0)
		disp_pos = 0;
}

	void
csrRight(n)
	int n;
{
	disp_pos += n;
	if (disp_pos > sizeof(disp_buf))
		disp_pos = sizeof(disp_buf);
}

/* delete N characters for regEdit*() */
	void
skkfep_delchars(n)
	int n;
{
	fep_delchars(n, FALSE);	/* FALSE:already moved cursor in regEdit*() func */
}

/* insert N characters for regEdit*() */
	int
skkfep_inschars(s)
	char *s;
{
	return fep_insstr(s);
}

	void
skkfep_update_screen()
{
	update_textscreen();
}

	void
rubout(n)
	int n;
{
	fep_delchars(n, TRUE);	/* TRUE:backspace */
	skkfep_update_screen();
}

	void
bell()
{
}

	void
flushOut(n)
	int n;
{
#ifdef REGMODE
	if (!Registering)	/* if Registering, output is to writes() */
#endif
		jpfep_flush = TRUE;	/* flush out fep_buf before print in writes() */
}

void
nulcmd(c)
char c;
{}


void
thru(c)
char c;
{
	writeShell1(c);
}


void
toAsc(c)
char c;
{
	flushKana();
	setKeymap(&CurrentKeymap,convertKeymap(&NormalKeymap));
	showSKKmode(SKK_MODE);
#ifdef REGMODE
	if (!Registering)		/* if Registering, must check ASCII, too */
#endif
		jpfep_mode = FALSE;	/* ASCII mode and quit jpfep_inject() */
}

void
toZenA(c)
char c;
{
	flushKana();
	setKeymap(&CurrentKeymap,ZenkakuKeymap);
	showSKKmode(ZENEI_MODE);
}

void
thruToAsc(c)
char c;
{
	toAsc(c);
	thru(c);
}

void
toEsc(c)
char c;
{
	flushKana();
	setKeymap(&CurrentKeymap,convertKeymap(&EscapedKeymap));
	showSKKmode(SKK_MODE);
}

void
thruToEsc(c)
char c;
{
	toEsc(c);
	thru(c);
}

void
thruBack(c)
char c;
{
	thru(c);
	toKana(c);
}

void
thru1(c)
char c;
{
	thru(c);
	restoreKeymap(&CurrentKeymap);
	showlastmode();
}

	static int
fep_insstr(s)
	char *s;
{
	int len = strlen(s);
	int disp_len = strlen(disp_buf);
	int i;

	for (i = disp_len; i >= disp_pos; i--) {	/* move to make space */
		if (i + len < sizeof(disp_buf))
			disp_buf[i + len] = disp_buf[i];
	}
	for (i = 0; i < len; i++) {					/* copy str on disp_pos */
		if (disp_pos + i >= sizeof(disp_buf)) {
			disp_buf[disp_pos + i - 1] = NUL;
			break;
		}
		disp_buf[disp_pos + i] = s[i];
	}
	disp_pos += len;
	if (disp_pos > sizeof(disp_buf))
		disp_pos = sizeof(disp_buf);
	update_textscreen();

	return len;
}

/* if change_pos is TRUE, works as Backspace('X')
 *   (move cursor left and delete('x') characters)
 * else only delete characters not changing Curpos.col.
 */
	static void
fep_delchars(len, change_pos)
	int len;
	int change_pos;	/* if TRUE, change disp_pos before delete characters */
{
	int i;

	if (change_pos)
		disp_pos -= len;
	
	if (disp_pos < 0)
		disp_pos = 0;
	for (i = 0; i <= strlen(&disp_buf[disp_pos + len]); i++) {
		if (disp_pos + len + i >= sizeof(disp_buf)) {
			disp_buf[disp_pos + i] = '\0';
			break;
		}
		if ((disp_buf[disp_pos + i] = disp_buf[disp_pos + len + i]) == '\0')
			break;
	}
	update_textscreen();
}

/* set SKKFEP keymap */
	static void
jpfep_setkeymap(skkmode)
	int skkmode;
{
	switch (skkmode) {
		case SKK_MODE:		/* main loop may have been ended by 'l' */
			toKana();
			break;
		case KANA_MODE:
		case KKANA_MODE:
			toKana();		/* Hira/Kata */
			break;
		case ZENEI_MODE:	/* toZenA() */
			setKeymap(&CurrentKeymap,ZenkakuKeymap);
			showSKKmode(skkmode);
			break;
		/* following cases may not occur */
		case KINPUT_MODE:	/* kkBeg() but do not WordBufLen = 0 */
			setKeymap(&CurrentKeymap,KanjiInputKeymap);
			break;
		case KALPHA_INPUT_MODE:	/* kkBegA() but do not WordBufLen = 0 */
			setKeymap(&CurrentKeymap,KAlphaInputKeymap);
			break;
		default:
			toKana();
			break;
	}
}

/* skkfep key event loop */
	static void
jpfep_inject()
{
	int c = NUL, o;

	disp_buf[disp_pos = 0] = NUL;

	if (jpfep_flush) {
		fep_insstr(delay_buf);
		delay_buf[0] = NUL;
		jpfep_flush = FALSE;
	}
	update_textscreen();

	/* SKKFEP main loop (get key input) */
	/* return injected Japanese characters as soon as possible
	 * ex: if typed "a" in Kana mode,
	 *     inject "a"(in kana) in fep_buf and return immediately.
	 */
	while (jpfep_mode && !jpfep_isready()) {
		o = OkuriFirst;
		OkuriFirst = 0;
		c = get_key(-1);
		if (IsKanji(c)) {
			thru(c);			/* inject into fep_buf */
			c = get_key(-1);	/* get second byte of kanji char */
			if (c >= 0)
				thru(c);
			break;
		}
		else if (c < 0) {
			continue;
		}
		else if (c == Ctrl('O') || c > 0x7f) {	/* end of IM */
			thru(c);
			break;
		}
		else	/* must be 0 <= c <= 0x7f for Keymap */
			(*CurrentKeymap[c])(c,o);
	}
	/* there is two ways to quit above main loop:
	 * 1) !jpfep_mode: typed 'l', or ESC when SimpleEsc mode.
	 * 2) jpfep_isready(): injected some Japanese characters in fep_buf.
	 */
	if (c == ESC) {	/* if typed ESC when jpfep_isready(), jpfep_mode != FALSE*/
		skkfep_curmode = skkfep_lastmode;	/* if SimpleEsc, curmode is Asc */
		jpfep_mode = FALSE;
	}
	else if (c == Ctrl('O')) {
		toAsc();
		skkfep_curmode = skkfep_lastmode;
	}
}

	static int
jpfep_init()
{
	char *env;
	if (ready)
		return ready;
	ready = TRUE;

	setEscape(SimpleEsc);
	LastEscapeBehavior = NoEsc;
	setKanaKey();
	/* to be able to change LearnMode on MS-DOS */
	if ((env = getenv("SKKFEP_LEARNMODE")) != NULL) {
		switch (*env) {
			case 'o': /* off */
			case 'n': /* none */
			case '0':
				NetLearnMode = LearnOff;
				break;
			case 't': /* temporal */
			case '1':
				NetLearnMode = LearnTemporal;
				break;
			case 'p':
			case '2':
				NetLearnMode = LearnPermanent;
				break;
			default:
				NetLearnMode = NET_LEARN_MODE;
				break;
		}
	}
	if (*UserDicName == '\0') {
		if ((env = getenv("HOME")) != NULL) {
			strcpy(UserDicName,env);
			strcat(UserDicName,PATHSEPSTR);
		}
		strcat(UserDicName,USER_DIC_NAME);
	}
	if (UserDic == NULL) {
		puts("Loading user SKK JISYO...");
		if ((UserDic = openSKK(UserDicName)) == NULL)
			puts("openSKK(UserDicName) failed.");
	}
#ifdef USE_SERVER
	if (ServerOK == 0) {
# if defined BOTH_SERVER_AND_SKKSRCH
		puts("Accessing global SKK JISYO(skkserv||SKK JISYO)");
# elif defined USE_SKKSRCH
		puts("Opening system SKK JISYO...");
# else
		puts("Connecting to skkserv...");
# endif
		if (openSKKserv() != -1)
		{
# ifdef BOTH_SERVER_AND_SKKSRCH
			openSKKservPtr = openSKKserv;
			getCandFromServerPtr = getCandFromServer;
			closeSKKservPtr = closeSKKserv;
# endif
			ServerOK = 1;
		}
# ifdef BOTH_SERVER_AND_SKKSRCH
		else if (openSKKsrch() != -1) {
			openSKKservPtr = openSKKsrch;
			getCandFromServerPtr = getCandFromSKKsrch;
			closeSKKservPtr = closeSKKsrch;
			ServerOK = 2;	/* use skksrch.c */
		}
# endif
	}
	if (ServerOK == 0)
		puts("Failed to access global SKK JISYO");
#endif
	return ready;
}

/* called before imskkfep.so removed */
void
_fini(void)
{
	if (!ready)
		return;
#ifdef USE_SERVER
	if (ServerOK) {
# ifdef BOTH_SERVER_AND_SKKSRCH
		closeSKKservPtr();
# else
		closeSKKserv();
# endif
		ServerOK = 0;
		ready = FALSE;
	}
#endif
#ifdef USERDIC_CHANGED_CHECK
	if (UserDic == NULL || !UserDic_Changed)
#else
	if (UserDic == NULL)
#endif
		return;
#ifdef USE_SERVER
	/* if any candidates were registered, 
	 * write UserDic regardless of NetLearnMode (even if LearnOff).
	 */
# if 0 /*ifdef USERDIC_CHANGED_CHECK {{if LearnOff, should not save for FD?}}*/
	if (NetLearnMode == LearnPermanent || (UserDic_Changed & UDIC_REGISTER))
# else
	if (NetLearnMode == LearnPermanent)
# endif
#endif
	{
		puts("Saving SKK JISYO...");
		closeSKK(UserDic,UserDicName);	/* save and free() UserDic */
		UserDic = NULL;
#ifdef USERDIC_CHANGED_CHECK
		UserDic_Changed = 0;
#endif
		ready = FALSE;
	}
}

/* called when imskkfep.so is loaded */
void
_init(void)
{
	jpfep_init();
	atexit(_fini);	/* {{ Is this OK? }} */
	jpfep_setkeymap(skkfep_curmode);	/* set keymap and show SKK mode */
	jpfep_mode = TRUE;
}

/*
 *	fep_modestring() return FEP mode string
 */
	char *
fep_modestring()
{
	/* use skkfep_lastshowmode instead of skkfep_curmode.
	 * This function is used only in showmode() to show SKK mode,
	 * so it is ok to return skkfep_lastshowmode only. (see showSKKmode())
	 */
	return modeString[skkfep_lastshowmode];
}

#include "mgl.h"
static struct textscreen *ts = NULL;

int
im_read(char *buf, int ibufsize)
{
	if (buf == NULL || ibufsize <= 0)
		return -1;

	if (!ready)
		_init();
	if (!jpfep_mode) {
		jpfep_mode = TRUE;
		jpfep_setkeymap(skkfep_curmode);	/* set keymap and show SKK mode */
	}
	if (!jpfep_isready()) {
		jpfep_inject();	/* pass key input to FEP and set fep_buf */
		/* here maybe !jpfep_isready() when only changed to ASCII mode */
		/* in such case, fall down and use vgetc() */
	}
	if (jpfep_isready()) {
		int len;
		if (fbtail - fbhead > ibufsize - 2) { /* -1: NUL, -1: to overwrite C-o*/
			len = ibufsize - 2;
			strncpy(buf, fbhead, len);
			buf[len] = NUL;
			fbhead += len;
		}
		else {
			len = fbtail - fbhead;
			strcpy(buf, fbhead);
			fbhead = fbtail = fep_buf;
			*fbtail = NUL;
		}
		if (!jpfep_mode && buf[len - 1] != Ctrl('O')) {
			buf[len] = Ctrl('O');
			buf[++len] = NUL;
		}
		if (ts != NULL) {
			free_textscreen(ts);
			refresh();
			ts = NULL;
		}
		return len;
	}
	buf[0] = Ctrl('O');
	buf[1] = NUL;
	if (ts != NULL) {
		free_textscreen(ts);
		refresh();
		ts = NULL;
	}
	return 1;
}


#define FONTSIZE	12
#define FONTWIDTH	(FONTSIZE / 2)	/* 6x12 font */
#define BOTTOMPOS	224	/* y position */
#define BORDERWIDTH	4
#define TOTALHEIGHT	((FONTSIZE) + (BORDERWIDTH) * 2)
#define TOPPOS		((BOTTOMPOS) - (TOTALHEIGHT))

static void
update_textscreen()
{
	char *modestr = fep_modestring();
	int len = strlen(modestr) + strlen(disp_buf) + 2;	/* +2: ' ' and cursor */

	if (ts != NULL) {
		free_textscreen(ts);
		ts = NULL;
	}

	ts = create_textscreen(NULL, 0, TOPPOS, len * FONTWIDTH + BORDERWIDTH * 2,
							TOTALHEIGHT, TS_SAVE|TS_BORDER|TS_BLINE);
	ts_clear(ts);
	set_font(FONTSIZE, 0);
	set_color(COLOR_BLACK);
	ts_put_string(ts, modestr, 0);
	ts_put_string(ts, " ", 0);
	ts_put_string(ts, disp_buf, 0);

	/* draw cursor marker */
	set_color(COLOR_REVERSE);
	fill_rect((strlen(modestr) + 1 + disp_pos) * FONTWIDTH + BORDERWIDTH,
		TOPPOS + BORDERWIDTH, (IsKanji(disp_buf[disp_pos])? 2: 1) * FONTWIDTH,
		FONTSIZE);
	refresh();
}
